import numpy as np
import operator
import itertools
import random
from deap import gp


# Define a protected division function
def protectedDiv(left, right):
    try:
        return left / right
    except ZeroDivisionError:
        return 1


# Define a new if-then-else function
def if_then_else(input, output1, output2):
    if input:
        return output1
    else:
        return output2


class Classifier:
    def __init__(self, data_manager):
        self.pset = None
        self.dm = data_manager
        """Dictionary used to rename arguments. Keys are the old names, values are the new names."""
        self.arg_mapper = {("IN" + str(i)): key for i, key in enumerate(self.dm.get_field_names())}
        """
        List of primitives to add to the functional set.
        Each value is a tuple of the form (<function>, [<parameter types>], <return type>).
        """
        self.primitives = [
            (operator.and_, [bool, bool], bool),
            (operator.or_, [bool, bool], bool),
            (operator.not_, [bool], bool),
            (operator.add, [float, float], float),
            (operator.sub, [float, float], float),
            (operator.mul, [float, float], float),
            (protectedDiv, [float, float], float),
            (operator.lt, [float, float], bool),
            (operator.gt, [float, float], bool),
            (operator.eq, [float, float], bool),
            (if_then_else, [bool, float, float], float),
            (operator.lt, [int, int], bool),
            (operator.gt, [int, int], bool),
            (operator.eq, [int, int], bool),
            (if_then_else, [bool, int, int], int),
        ]
        terminals = list([(i, int) for i in np.unique(self.dm.get_data_set()[['protocol_type', 'service', 'flag']])])
        terminals += [(False, bool), (True, bool)]
        """
        List of terminals to add to the terminal set.
        Each value is a tuple of the form (<value>, <data type>).
        """
        self.terminal_set = terminals
        """A list of data types representing the data types for each input value of a record."""
        self.input_types = [float, int, int, int] + list(itertools.repeat(float, len(self.dm.get_field_names()) - 5))
        """The final return type of the classifier."""
        self.return_type = bool

    def create_structure(self):
        """
        Creates the primitive set for this classifier by making use of strongly typed GP.
        :return: self
        """
        self.pset = gp.PrimitiveSetTyped("MAIN", self.input_types, self.return_type, "IN")
        self.rename_arguments()
        self.load_primitives()
        self.pset.addEphemeralConstant("rand100", lambda: random.random() * 100, float)
        self.load_terminal_set()
        return self

    def rename_arguments(self):
        """
        Renames the arguments in the structure.
        :return: self
        """
        if self.arg_mapper:
            self.pset.renameArguments(**self.arg_mapper)
        return self

    def load_ephemerals(self):

        return self

    def load_terminal_set(self):
        """
        Adds the terminals of the classifier to the terminal set of the classifier's primitive set.
        :return: self
        """
        for t in self.terminal_set:
            self.pset.addTerminal(*t)
        return self

    def load_primitives(self):
        """
        Loads the primitives of this classifier into the functional set.
        :return:
        """
        for p in self.primitives:
            self.pset.addPrimitive(*p)
        return self

    def get_structure(self):
        """
        Gets the primitive set of the classifier.
        :return: The primitive set.
        """
        return self.pset



import pandas as pd
import numpy as np
from sklearn.preprocessing import LabelEncoder


class DataManager:
    def __init__(self):
        self.dataset = pd.DataFrame()
        self.training_set = pd.DataFrame()
        self.validation_set = pd.DataFrame()
        self.testing_set = pd.DataFrame()
        self.fieldnames = []

    def load_dataset(self):
        """
        Loads the data set from the files and formats it. This function is responsible for setting the fieldnames
        and the dataset properties of the data manager.
        :return: self
        """
        # Read the different field names
        self.fieldnames = pd.read_csv('field_names.csv', header=None, usecols=[0], squeeze=True).tolist()
        # Read the data, with the appropriate headings
        dataset = pd.read_csv('data_set.csv', header=None, names=self.fieldnames)
        replacements = [False if t in ['unknown', 'normal'] else True for t in dataset.attack_type.unique()]
        dataset.attack_type.replace(dataset.attack_type.unique(), replacements, inplace=True)

        self.fieldnames = self.fieldnames[:-1]
        dataset.drop(columns=['difficulty_level'], inplace=True)

        le = LabelEncoder()
        dataset.protocol_type = le.fit_transform(dataset.protocol_type)
        dataset.service = le.fit_transform(dataset.service)
        dataset.flag = le.fit_transform(dataset.flag)

        dataset[dataset.columns.difference(['attack_type'])] = dataset[
            dataset.columns.difference(['attack_type'])].astype(float)
        dataset[['protocol_type', 'service', 'flag']] = dataset[['protocol_type', 'service', 'flag']].astype(int)
        self.dataset = dataset
        return self

    def split_dataset(self, train_ratio, test_ratio):
        """
        Splits the loaded data set into a training, validation and testing set based on the given ratios.
        :param train_ratio: The ratio of the training set in relation to the size of the data set.
        :param test_ratio: The ratio of the testing set in relation to the size of the data set.
        :return: self
        """
        self.training_set, self.testing_set, self.validation_set = np.split(
            self.dataset.sample(frac=1.0),
            [int(train_ratio * len(self.dataset)), int((train_ratio + test_ratio) * len(self.dataset))])
        return self

    def get_training_set(self):
        return self.training_set

    def get_testing_set(self):
        return self.testing_set

    def get_validation_set(self):
        return self.validation_set

    def get_data_set(self):
        return self.dataset

    def get_field_names(self):
        return self.fieldnames



import random
import operator
import pickle
import datetime
import numpy

from deap import algorithms
from deap import base
from deap import creator
from deap import tools
from deap import gp

from geepers.data_manager import DataManager
from geepers.classifier import Classifier


class GP:
    def __init__(self, population_size, num_gens, crossover_prob, mutation_prob, tournament_size, parsimony_size, max_height=None, sample_frac=1.0):
        self.population_size = population_size
        self.num_gens = num_gens
        self.crossover_probability = crossover_prob
        self.mutation_probability = mutation_prob
        self.tournament_size = tournament_size
        self.parsimony_size = parsimony_size
        self.max_height = max_height
        self.sample_frac = sample_frac

        self.dm = DataManager()
        self.dm.load_dataset()
        definer = Classifier(data_manager=self.dm)
        definer.create_structure()
        self.pset = definer.get_structure()
        creator.create("FitnessMax", base.Fitness, weights=(1.0,))
        creator.create("Individual", gp.PrimitiveTree, fitness=creator.FitnessMax)
        self.toolbox = base.Toolbox()

    def get_accuracy(self, individual, records):
        func = self.toolbox.compile(expr=individual)
        results = [bool(func(*record[:-1])) for record in records.values]
        result = len(records[records.attack_type == results])
        return result

    def classify(self, individual):
        return self.get_accuracy(individual, self.dm.get_training_set().sample(frac=self.sample_frac)),

    def init_toolbox(self):
        toolbox = self.toolbox
        pset = self.pset
        toolbox.register("expr", gp.genHalfAndHalf, pset=pset, min_=1, max_=5)
        toolbox.register("individual", tools.initIterate, creator.Individual, toolbox.expr)
        toolbox.register("population", tools.initRepeat, list, toolbox.individual)
        toolbox.register("compile", gp.compile, pset=pset)
        toolbox.register("evaluate", self.classify)
        toolbox.register("select", tools.selDoubleTournament, fitness_size=self.tournament_size, parsimony_size=self.parsimony_size, fitness_first=True)
        toolbox.register("mate", gp.cxOnePoint)
        toolbox.register("expr_mut", gp.genHalfAndHalf, min_=0, max_=2)
        toolbox.register("mutate", gp.mutUniform, expr=toolbox.expr_mut, pset=pset)
        toolbox.decorate("mate", gp.staticLimit(operator.attrgetter('height'), self.max_height))
        toolbox.decorate("mutate", gp.staticLimit(operator.attrgetter('height'), self.max_height))
        return self

    def run(self):
        pop = self.toolbox.population(n=self.population_size)
        hof = tools.HallOfFame(1)
        stats = tools.Statistics(lambda ind: ind.fitness.values)
        stats.register("avg", numpy.mean)
        stats.register("std", numpy.std)
        stats.register("min", numpy.min)
        stats.register("max", numpy.max)

        print('Crossover Rate: {}, Mutation Rate: {}'.format(self.crossover_probability, self.mutation_probability))

        final_pop, logbook = algorithms.eaSimple(pop, self.toolbox, cxpb=self.crossover_probability,
                                                 mutpb=self.mutation_probability, ngen=self.num_gens, stats=stats, halloffame=hof,
                                                 verbose=True)

        return pop, stats, hof, final_pop, logbook, self.crossover_probability, self.mutation_probability


if __name__ == "__main__":
    for i in range(0, 10):
        try:
            print()
            print("Starting run #{}...".format(i))
            start_time = datetime.datetime.now()
            print("Start time: {}".format(str(start_time)))
            runner = GP(250, 15, random.uniform(0.0, 1.0), random.uniform(0.0, 1.0), 50, 1.4, 10, 0.2)
            runner.dm.split_dataset(0.6, 0.2)
            runner.init_toolbox()
            pop, stats, hof, final_pop, logbook, crossover_probability, mutation_probability = runner.run()
            execution_time = datetime.datetime.now() - start_time
            print("Finished, saving data...")
            print("End time: {}, Execution Time: {}".format(str(datetime.datetime.now()), str(execution_time)))
            saved_data = {
                "hof": hof,
                "logbook": logbook,
                "population": final_pop,
                "crossover_probability": crossover_probability,
                "mutation_probability": mutation_probability,
            }
            with open("classifiers/{}.pkl".format(i), "wb") as save_file:
                pickle.dump(saved_data, save_file)
            training_set, validation_set, testing_set = runner.dm.get_training_set(), runner.dm.get_validation_set(), runner.dm.get_testing_set()
            print("Training Accuracy: {}".format(runner.get_accuracy(hof[0], training_set) / len(training_set) * 100))
            print("Validation Accuracy: {}".format(runner.get_accuracy(hof[0], validation_set) / len(validation_set) * 100))
            print("Testing Accuracy: {}".format(runner.get_accuracy(hof[0], testing_set) / len(testing_set) * 100))
            print()
        except Exception as ex:
            print()
            print("=" * 50)
            print('UNEXPECTED ERROR OCCURRED DURING RUN #{}'.format(i))
            print(str(ex))
            print("=" * 50)
            print()



import random
import pickle
import datetime
import numpy
import array
from deap import algorithms
from deap import base
from deap import creator
from deap import tools
from deap import gp
from geepers.data_manager import DataManager
from geepers.classifier import Classifier


class HyperHeuristic:
    def __init__(self, population_size, num_gens, crossover_prob, mutation_prob, chromosome_length, tournament_size, sample_size):
        self.population_size = population_size
        self.num_gens = num_gens
        self.crossover_probability = crossover_prob
        self.mutation_probability = mutation_prob
        self.chromosome_length = chromosome_length
        self.tournament_size = tournament_size
        self.sample_size = sample_size

        self.dm = DataManager()
        self.dm.load_dataset()
        creator.create("FitnessMax", base.Fitness, weights=(1.0,))
        creator.create("Individual", gp.PrimitiveTree, fitness=creator.FitnessMax)
        definer = Classifier(data_manager=self.dm)
        definer.create_structure()
        self.classifier_pset = definer.get_structure()
        toolbox = base.Toolbox()
        toolbox.register("compile", gp.compile, pset=self.classifier_pset)
        self.toolbox = toolbox
        self.classifiers = []

    def load_classifiers(self):
        classifiers = []
        for i in range(20):
            with open('classifiers/{}.pkl'.format(i), 'rb') as file:
                data = pickle.load(file)
            classifiers.append(data)

        classifiers = sorted(classifiers, key=lambda x: x['hof'][0].fitness, reverse=True)[:5]
        for c in classifiers:
            c['func'] = self.toolbox.compile(expr=c['hof'][0])
        self.classifiers = classifiers
        return self

    def init_toolbox(self):
        creator.create("Combo", array.array, typecode='i', fitness=creator.FitnessMax)
        self.toolbox.register('hh_sample', random.randrange, 0, len(self.classifiers))
        self.toolbox.register('individual', tools.initRepeat, creator.Combo, self.toolbox.hh_sample, self.chromosome_length)
        self.toolbox.register('population', tools.initRepeat, list, self.toolbox.individual)
        self.toolbox.register('evaluate', self.eval_classification)
        self.toolbox.register('mate', tools.cxTwoPoint)
        self.toolbox.register('mutate', tools.mutUniformInt, low=0, up=len(self.classifiers) - 1, indpb=0.5)
        self.toolbox.register('select', tools.selTournament, tournsize=self.tournament_size)

    def get_accuracy(self, individual, records):
        results = [bool(self.classifiers[individual[index % len(individual)]]['func'](*record[:-1]))
                   for index, record in enumerate(records.values)]
        result = len(records[records.attack_type == results])
        return result

    def eval_classification(self, individual):
        records = self.dm.get_training_set().sample(n=self.sample_size)
        return self.get_accuracy(individual, records),

    def run(self):
        pop = self.toolbox.population(n=self.population_size)
        hof = tools.HallOfFame(1)
        stats = tools.Statistics(lambda ind: ind.fitness.values)
        stats.register("avg", numpy.mean)
        stats.register("std", numpy.std)
        stats.register("min", numpy.min)
        stats.register("max", numpy.max)

        print('Crossover Rate: {}, Mutation Rate: {}'.format(self.crossover_probability, self.mutation_probability))
        pop, log = algorithms.eaSimple(pop, self.toolbox, cxpb=self.crossover_probability, mutpb=self.mutation_probability,
                                       ngen=self.num_gens, stats=stats, halloffame=hof, verbose=True)

        return pop, log, hof, self.crossover_probability, self.mutation_probability


if __name__ == "__main__":
    for i in range(0, 5):
        try:
            print()
            print("Starting run #{}...".format(i))
            start_time = datetime.datetime.now()
            print("Start time: {}".format(str(start_time)))
            hh = HyperHeuristic(250, 30, random.uniform(0.0, 1.0), random.uniform(0.0, 1.0), 10, 10, 10000)
            hh.dm.split_dataset(0.6, 0.2)
            hh.load_classifiers()
            hh.init_toolbox()
            pop, logbook, hof, crossover_probability, mutation_probability = hh.run()
            execution_time = datetime.datetime.now() - start_time
            print("Finished, saving data...")
            print("End time: {}, Execution Time: {}".format(str(datetime.datetime.now()), str(execution_time)))
            saved_data = {
                "hof": hof,
                "logbook": logbook,
                "population": pop,
                "crossover_probability": crossover_probability,
                "mutation_probability": mutation_probability,
            }
            with open("hyper_heuristics/{}.pkl".format(i), "wb") as save_file:
                pickle.dump(saved_data, save_file)
            training_set, validation_set, testing_set = hh.dm.get_training_set(), hh.dm.get_validation_set(), hh.dm.get_testing_set()
            print("Training Accuracy: {}".format(hh.get_accuracy(hof[0], training_set) / len(training_set) * 100))
            print("Validation Accuracy: {}".format(hh.get_accuracy(hof[0], validation_set) / len(validation_set) * 100))
            print("Testing Accuracy: {}".format(hh.get_accuracy(hof[0], testing_set) / len(testing_set) * 100))
            print()
        except Exception as ex:
            print()
            print("=" * 50)
            print('UNEXPECTED ERROR OCCURRED DURING RUN #{}'.format(i))
            print(str(ex))
            print("=" * 50)
            print()

